package tech.graphits.catalogue;

import com.google.common.base.Charsets;
import com.google.common.io.Resources;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.neo4j.driver.v1.GraphDatabase;
import org.neo4j.ogm.config.ClasspathConfigurationSource;
import org.neo4j.ogm.config.Configuration;
import org.neo4j.ogm.config.ConfigurationSource;
import org.neo4j.ogm.cypher.Filter;
import org.neo4j.ogm.cypher.Filters;
import org.neo4j.ogm.cypher.query.Pagination;
import org.neo4j.ogm.cypher.query.SortOrder;
import org.neo4j.ogm.driver.Driver;
import org.neo4j.ogm.drivers.bolt.driver.BoltDriver;
import org.neo4j.ogm.model.QueryStatistics;
import org.neo4j.ogm.model.Result;
import org.neo4j.ogm.session.Session;
import org.neo4j.ogm.session.SessionFactory;
import org.neo4j.ogm.transaction.Transaction;
import tech.graphits.catalogue.model.APourTheme;
import tech.graphits.catalogue.model.Livre;
import tech.graphits.catalogue.model.Personne;

import java.io.IOException;
import java.util.*;
import java.util.concurrent.*;
import java.util.stream.Collectors;

import static org.hamcrest.CoreMatchers.hasItems;
import static org.junit.Assert.*;
import static org.neo4j.ogm.cypher.ComparisonOperator.EQUALS;

public class CatalogueTest {

    private Session session;
    private SessionFactory sessionFactory;

    @Before
    public void setup() throws IOException {
        sessionFactory = new SessionFactory("tech.graphits.catalogue.model");
        session = sessionFactory.openSession();

        session.purgeDatabase();
        session.query(Resources.toString(Resources.getResource("test_data.cql"), Charsets.UTF_8), Collections.emptyMap());
    }

    @After
    public void rollback() {
        session.clear();
    }

    public void initAvecProperties() {

        // tag::init_avec_properties[]
        ConfigurationSource props = new ClasspathConfigurationSource("ogm.properties");
        Configuration config = new Configuration.Builder(props).build();
        sessionFactory = new SessionFactory(config, "tech.graphits.catalogue.model");
        // end::init_avec_properties[]
    }

    public void initProgrammatique() {

        // tag::init_programmatique[]
        Configuration config = new Configuration.Builder()
            .uri("bolt://monServeur")
            .credentials("utilisateur", "mot de passe")
            .build();
        sessionFactory = new SessionFactory(config, "tech.graphits.catalogue.model");
        // end::init_programmatique[]
    }

    public void initAvecDriver() {

        // tag::init_avec_driver[]
        org.neo4j.driver.v1.Driver nativeDriver = GraphDatabase.driver("bolt://monServeur");
        Driver ogmDriver = new BoltDriver(nativeDriver);
        sessionFactory = new SessionFactory(ogmDriver, "tech.graphits.catalogue.model");
        // end::init_avec_driver[]
    }

    @Test
    public void recherche_avec_queryForObject() {
        // la recherche ne fonctionne pas pcq on crée une nouvelle factory avec une nouvelle bdd embarquée
        // pas grave c'est pour l'exemple

        Configuration config = new Configuration.Builder().build();
        // tag::recherche_avec_queryForObject[]
        SessionFactory sessionFactory = new SessionFactory(config, "tech.graphits.catalogue.model"); // <1>
        Session session = sessionFactory.openSession();

        Personne sylvain = session.queryForObject(Personne.class, // <2>
                "match (p:Personne{email:{email}}) return p", // <3>
                Collections.singletonMap("email", "sylvain.roussy@graphits.tech")); // <4>

        assertEquals("Sylvain Roussy", sylvain.getNom());
        assertEquals("sylvain.roussy@graphits.tech", sylvain.getEmail());
        // end::recherche_avec_queryForObject[]

        // ne charge pas les entités dépendantes
        assertEquals(0, sylvain.getLivres().size());
    }

    @Test
    public void recherche_avec_loadAll() {

        // tag::recherche_avec_loadAll[]
        Filter emailFilter = new Filter("email", EQUALS, "sylvain.roussy@graphits.tech");
        Collection<Personne> results = session.loadAll(
                Personne.class // <1>
                , new Filters(emailFilter) // <2>
                , 1); // <3>

        Personne result = results.iterator().next();
        assertEquals("Sylvain Roussy", result.getNom());
        assertEquals("sylvain.roussy@graphits.tech", result.getEmail());

        // charge les entités dépendantes de niveau 1
        assertEquals(1 , result.getLivres().size());
        Livre livre = result.getLivres().iterator().next();
        assertEquals("Neo4j - Des données et des graphes : Tome 2 - Déploiement", livre.getTitre());

        // charge le niveau 1 mais pas le niveau 2
        assertNull(livre.getProduit());
        // end::recherche_avec_loadAll[]
    }

    @Test
    public void recherche_avec_query() {

        // tag::recherche_avec_query[]
        Result results = session.query("MATCH (p:Personne) --> (l:Livre) RETURN p.nom as nom, count(l) as nbLivres", Collections.emptyMap());
        // tag::recherche_avec_query[]

        Map<Object, Object> allResults = new HashMap<>();
        results.forEach(map -> allResults.put(map.get("nom"), map.get("nbLivres")));
        assertEquals(3, allResults.size());
        assertThat(allResults.keySet(), hasItems("Nicolas Mervaillie", "Nicolas Rouyer", "Sylvain Roussy"));
    }

    @Test
    public void mise_a_jour_simple() {

        long nbPersonnes = session.countEntitiesOfType(Personne.class);

        Personne personne = new Personne();
        personne.setNom("Brian");
        personne.setEmail("brian@graphit.tech");

        session.save(personne);

        assertEquals(nbPersonnes + 1, session.countEntitiesOfType(Personne.class));
    }

    @Test
    public void sauvegarde_en_cascade() {

        // tag::creation[]
        session.purgeDatabase();

        Livre livre = new Livre();
        livre.setTitre("un titre");
        livre.setTitreMaitre("un titre maitre");
        livre.setSousTitre("un sous titre");

        Personne auteur = new Personne();
        auteur.setEmail("brian@graphits.tech");
        auteur.setNom("Brian");
        auteur.setLivres(Collections.singleton(livre));

        session.save(auteur);

        assertEquals(1, session.countEntitiesOfType(Personne.class));
        assertEquals(1, session.countEntitiesOfType(Livre.class));

        Filter filtreTitre = new Filter("titre", EQUALS, "un titre");
        Livre nouveauLivre = session.loadAll(Livre.class, new Filters(filtreTitre)).iterator().next();
        assertEquals("un titre", nouveauLivre.getTitre());
        // end::creation[]

        // tag::maj[]
        livre.setSousTitre("un nouveau sous titre");

        session.save(auteur);

        nouveauLivre = session.loadAll(Livre.class, new Filters(filtreTitre)).iterator().next();
        assertEquals("un nouveau sous titre", nouveauLivre.getSousTitre());
        // end::maj[]
    }

    @Test
    public void mise_a_jour_en_cascade() {

        Filter emailFilter = new Filter("email", EQUALS, "sylvain.roussy@graphits.tech");
        Personne sylvain = session.loadAll(Personne.class, new Filters(emailFilter)).iterator().next();

        // on change le nom
        final String NOUVEL_EMAIL = "sylvain@graphits.tech";
        sylvain.setEmail(NOUVEL_EMAIL);
        // on change le titre du livre deja rattache
        sylvain.getLivres().iterator().next().setSousTitre("un test");
        // on ajoute un livre
        Livre tome1 = new Livre("Neo4j - Des données et des graphes : Tome 1 - Prise en main", "Neo4j - Des données et des graphes", "Tome 1 - Prise en main");
        sylvain.getLivres().add(tome1);

        session.save(sylvain);

        sylvain = session.load(Personne.class, sylvain.getId());
        assertEquals(NOUVEL_EMAIL, sylvain.getEmail());
        assertEquals(2, sylvain.getLivres().size());

        Set<String> titres = sylvain.getLivres().stream().map(Livre::getTitre).collect(Collectors.toSet());
        assertTrue(titres.contains("Neo4j - Des données et des graphes : Tome 1 - Prise en main"));
    }

    @Test
    public void relation_porteuse_de_donnees() {

        Filter filtreTitreMaitre = new Filter("titreMaitre", EQUALS, "Neo4j - Des données et des graphes");
        Livre livre = session.loadAll(Livre.class, new Filters(filtreTitreMaitre))
            .iterator().next();

        Set<String> categories = livre.getaPourTheme().stream()
                .map(aPourTheme -> aPourTheme.getCategorie().getTitre())
                .collect(Collectors.toSet());
        assertThat(categories, hasItems("Base de données", "Programmation"));

        Set<Double> proximites = livre.getaPourTheme().stream()
                .map(APourTheme::getProximite)
                .collect(Collectors.toSet());
        assertThat(proximites, hasItems(1.0, 0.3));
    }

    @Test
    public void exemple_transaction() {

        // tag::transaction[]
        Filter emailFilter = new Filter("email", EQUALS, "sylvain.roussy@graphits.tech");
        try (Transaction tx = session.beginTransaction()) { // <1>
            Personne sylvain = session.loadAll(Personne.class,
                    new Filters(emailFilter)).iterator().next();
            sylvain.setNom("Sylvain R");
            tx.rollback(); // <2>
            // ou tx.commit()
        }
        // end::transaction[]
    }

    public void lecture_seule() {

        // tag::lecture_seule[]
        try (Transaction tx = session.beginTransaction(Transaction.Type.READ_ONLY)) {
            // accès aux données
            tx.commit();
        }
        // end::lecture_seule[]
    }

    public void exemple_bookmarks() {

        // tag::bookmarks[]
        try (Transaction tx = session.beginTransaction()) { // <1>
            // accès base
            tx.commit();
        }
        String bookmark = session.getLastBookmark();

        // la session est clôturée, le bookmark est retourné au client
        // du temps passe...
        // le client réalise un nouvel appel, en passant le bookmark en paramètre

        session.withBookmark(bookmark);
        // les accès aux données seront alors garantis consistants avec les mises à jour précédentes
        // end::bookmarks[]
    }

    @Test
    public void transaction_simple() {

        Filter emailFilter = new Filter("email", EQUALS, "sylvain.roussy@graphits.tech");
        Transaction tx = session.beginTransaction();
        try {
            Personne sylvain = session.loadAll(Personne.class,
                    new Filters(emailFilter)).iterator().next();
            sylvain.setNom("Sylvain R");
            uneMethodeQuiLeveUneException();
            tx.commit();
        } catch (Exception e) {
            tx.rollback();
        }
        assertEquals(Transaction.Status.ROLLEDBACK, tx.status());
        session = sessionFactory.openSession(); // <1>
        Personne sylvain = session.loadAll(Personne.class,
                new Filters(emailFilter)).iterator().next();
        assertEquals("Sylvain Roussy", sylvain.getNom());
    }

    private void uneMethodeQuiLeveUneException() throws Exception {
        throw new Exception("test");
    }

    @Test
    public void exemple_aggregation() {

        // tag::aggregation[]
        Result result = session.query("MATCH (p:Personne) -[:EST_AUTEUR]-> (l:Livre) "+
                "RETURN l.titreMaitre as titre, collect(p.nom) as auteurs "
                , Collections.emptyMap()); // <1>

        Map<String, Object> record = result.queryResults().iterator().next(); // <2>

        assertEquals("Neo4j - Des données et des graphes", record.get("titre"));
        String[] auteurs = (String[]) record.get("auteurs");
        Arrays.sort(auteurs);
        assertArrayEquals(new String[]{"Nicolas Mervaillie", "Nicolas Rouyer", "Sylvain Roussy"}
                , auteurs);
        // end::aggregation[]
    }

    @Test
    public void exemple_pagination() {

        // tag::pagination[]
        Pagination paging = new Pagination(0, 10);
        SortOrder order = new SortOrder().add(SortOrder.Direction.ASC, "nom");

        Iterable<Personne> auteurs = session.loadAll(Personne.class, order, paging);
        // end::pagination[]

        Iterator<Personne> iterator = auteurs.iterator();
        assertEquals("Nicolas Mervaillie", iterator.next().getNom());
        assertEquals("Nicolas Rouyer", iterator.next().getNom());
        assertEquals("Sylvain Roussy", iterator.next().getNom());
    }

    @Test
    public void exemple_update_cypher() {

        // tag::update_cypher[]
        Result result = session.query("MATCH (l:Livre) SET l.titre = l.titreMaitre + l.sousTitre"
                , Collections.emptyMap());

        QueryStatistics queryStatistics = result.queryStatistics();

        assertEquals(1, queryStatistics.getPropertiesSet());
        assertEquals(0, queryStatistics.getNodesCreated());
        // end::update_cypher[]
    }


    @Test
    public void des_tx_ne_doivent_pas_pouvoir_modifier_le_meme_noeud() throws InterruptedException, ExecutionException {

        CountDownLatch latch = new CountDownLatch(1);

        // on démarre une tx et on commence à modifier un noeud
        Transaction transaction = session.beginTransaction();
        Filter emailFilter = new Filter("email", EQUALS, "sylvain.roussy@graphits.tech");
        Personne sylvain = session.loadAll(Personne.class, new Filters(emailFilter)).iterator().next();
        sylvain.setNom("sylvain1");
        session.save(sylvain);

        // dans un autre thread et une autre tx, on essaie de modifier le même noeud
        ExecutorService pool = Executors.newFixedThreadPool(1);
        pool.submit(() -> {
            Session session2 = sessionFactory.openSession();
            Transaction transaction2 = session2.beginTransaction();
            assertNotEquals(transaction, transaction2);
            Personne sylvain2 = session2.loadAll(Personne.class, new Filters(emailFilter)).iterator().next();
            sylvain2.setNom("sylvain2");
            latch.countDown();
            // reste bloqué, car tx deja en cours sur la même noeud
            session2.save(sylvain2);
            transaction2.commit();
            return 0;
        });

        // on attend le demarrage de l'autre tx
        latch.await();
        Thread.sleep(100); // attend save

//        Personne p1 = Executors.newSingleThreadExecutor().submit(() -> {
//            Session sessionx = sessionFactory.openSession();
//            return sessionx.loadAll(Personne.class, new Filters("email", "sylvain.roussy@graphits.tech")).iterator().next();
//        }).get();
//        assertEquals("Sylvain Roussy", p1.getNom());
        // verifie que le save est bloqué, que la maj du prenom dans la 2eme tx n'est pas visible ici
        Personne p1 = session.loadAll(Personne.class, new Filters(emailFilter)).iterator().next();
        assertEquals("sylvain1", p1.getNom());

        // fin de la tx, la 2eme tx doit maintenant pouvoir continuer
        transaction.commit();

        // on verifie que la 2eme tx s'est terminé correctement
        pool.shutdown();
        boolean termine = pool.awaitTermination(100, TimeUnit.MILLISECONDS);
        assertTrue(termine);

        // les modifs de la 2eme tx sont maintenant visibles
        Personne saved = sessionFactory.openSession().loadAll(Personne.class, new Filters(emailFilter)).iterator().next();
        assertEquals("sylvain2", saved.getNom());
    }
}
