package tech.graphits.inserter;

import org.neo4j.graphdb.Label;
import org.neo4j.graphdb.RelationshipType;
import org.neo4j.io.fs.FileUtils;
import org.neo4j.unsafe.batchinsert.BatchInserter;
import org.neo4j.unsafe.batchinsert.BatchInserters;

import java.io.*;
import java.nio.charset.StandardCharsets;
import java.util.HashMap;
import java.util.Map;



public class LivreInserter {

    private Map<String,Auteur> auteurs;
    private Map<String,Livre> livres;
    private Map<String,Long> auteurNodes, livreNodes;
    
    private String cheminVersDb = "livres.graphdb";
    
    private BatchInserter batchInserter;

    public LivreInserter (String cheminVersDb){
    	
    	this.cheminVersDb = cheminVersDb;
    	
    	auteurNodes = new HashMap<String,Long>();
        livreNodes = new HashMap<String,Long>();

        auteurs = new HashMap<String,Auteur>();
        livres = new HashMap<String, Livre>();
    }
    
    /**
     * Patron de mthodes
     * @throws IOException
     */
    public void execute() throws IOException {
    	
    	this.initialisationBatchInserter();
    	this.chargeLivres();
    	this.chargeAuteurs();
    	this.insert();
    	this.shutdown();
    }
    
    /**
     * Instancie l'objet BatchInserter qui pointe sur les fichiers du graphe
     * @return l'objet BatchInserter
     * @throws IOException
     */
    private void initialisationBatchInserter () throws IOException{   	
       final File importedData = new File( cheminVersDb );
       
       // rinitialisation du graphe (attention, perte des donnes)
        FileUtils.deleteRecursively(importedData);
       
       batchInserter = BatchInserters.inserter(importedData.getAbsoluteFile());
      
   }
    
    /**
     * Effectue une transformation de l'objet livre en dictionnaire
     * @param livre
     * @return dictionnaire
     */
    private Map<String, Object> livreToProperty(Livre livre) {
        final Map<String, Object> properties = new HashMap<String, Object>();

        properties.put("titre", livre.getTitre());
        properties.put("titreMaitre", livre.getTitreMaitre());
        properties.put("sousTitre",livre.getSousTitre());

        return properties;
    }

    /**
     * Effectue une transformation de l'objet auteur en dictionnaire
     * @param auteur
     * @return dictionnaire
     */
    private Map<String, Object> auteurToProperty(Auteur auteur) {
       final  Map<String, Object> properties = new HashMap<String, Object>();
       properties.put("email", auteur.getEmail());
       return properties;
    }
    
    private void shutdown (){
    	batchInserter.shutdown();
    	auteurNodes.clear();
        livreNodes.clear();

        auteurs.clear();
        livres.clear();
        System.out.println("Dchargement termin");
    }
   
    private void chargeAuteurs () throws UnsupportedEncodingException, FileNotFoundException, IOException{
    	  BufferedReader fileReader = null;
    	  
    	  try {
    		  fileReader = new BufferedReader(
	                  new InputStreamReader(
	                		  ClassLoader.getSystemResourceAsStream("auteurs.csv"), StandardCharsets.UTF_8.name()));
	          String line;
	          boolean entete = true;
	          // Lecture du fichier ligne par ligne
	          while ((line = fileReader.readLine()) != null){
	        	  
	        	  // Saut de la premire ligne
	        	  if (entete){
	        		  entete = false;
	        		  continue;
	        	  }
	            
	              System.out.println("LivreInserter - line = " + line);
	             
	              final String[] a = line.split(";", -1);	
	             
	              System.out.println("LivreInserter - a.size = " + a.length);
	
	              auteurs.put(a[0],new Auteur(a[0]));
	          }
    	  }
    	  finally {
    		  if (fileReader != null) fileReader.close();
    	  }
    }
    
    
    private void chargeLivres () throws UnsupportedEncodingException, FileNotFoundException, IOException{
  	  BufferedReader fileReader = null;
  	  
  	  try {
  		  fileReader = new BufferedReader(
	                  new InputStreamReader(
	                		  ClassLoader.getSystemResourceAsStream("livres.csv"),
	                		  StandardCharsets.UTF_8.name()));
	          String line;	      
	          boolean entete = true;
	          // Lecture du fichier ligne par ligne
	          while ((line = fileReader.readLine()) != null){
	            
	        	  // Saut de la premire ligne
	        	  if (entete){
	        		  entete = false;
	        		  continue;
	        	  }
	        	  
	              System.out.println("LivreInserter - line = " + line);
	            
	              final String[] l = line.split(";", -1);
	
	              System.out.println("LivreInserter - l.size = " + l.length);

	              livres.put(l[0],new Livre(l[0], l[1], l[2]));
	          }
  	  }
  	  finally {
  		  if (fileReader != null) fileReader.close();
  	  }
  }

    
    
    private void insert () throws UnsupportedEncodingException, FileNotFoundException, IOException{        

        // Cration des labels
        Label auteurLabel = Label.label("Auteur");
        Label livreLabel = Label.label("Livre");
        batchInserter.createDeferredSchemaIndex( auteurLabel ).on("email").create();
        batchInserter.createDeferredSchemaIndex( livreLabel ).on("titre").create();

        // cration du type de relation EST_AUTEUR
        RelationshipType estAuteur =  RelationshipType.withName("EST_AUTEUR");

        // cration des noeuds Auteur
        for (Auteur auteur : auteurs.values()) {
            long node = batchInserter.createNode(auteurToProperty(auteur), auteurLabel);
            auteurNodes.put(auteur.getEmail(), node);

        }

        // cration des noeuds Livre
        for (Livre livre : livres.values()) {
            long node = batchInserter.createNode(livreToProperty(livre), livreLabel);
            livreNodes.put(livre.getTitre(), node);

        }

        // cration de la relation auteur => livre (EST_AUTEUR)
        BufferedReader fileReader = new BufferedReader(
                new InputStreamReader(
                		ClassLoader.getSystemResourceAsStream("auteurs_livres.csv"), "UTF8"));
        // Lecture du fichier ligne par ligne
        String line ="";
        while ((line = fileReader.readLine()) != null)
        {
            // On prend tous les tokens disponibles
            String[] auteurslivres = line.split(";", -1);

            String email = auteurslivres[0];
            String titre = auteurslivres[1];

            if ( auteurNodes.containsKey(email) ) {
                if ( livreNodes.containsKey(titre) ) {
                    // cration de la relation EST_AUTEUR
                    Map<String, Object> estAuteurProperties = new HashMap<String, Object>();
                    batchInserter.createRelationship(
                        auteurNodes.get(email),
                        livreNodes.get(titre),
                        estAuteur,
                        estAuteurProperties);

                }
            }
        }
        System.out.println("Insertions termines");

      
    }

}