package tech.graphits.catalogue;

import java.io.IOException;

import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.neo4j.driver.v1.AuthToken;
import org.neo4j.driver.v1.AuthTokens;
import org.neo4j.driver.v1.Driver;
import org.neo4j.driver.v1.GraphDatabase;
import org.neo4j.driver.v1.Record;
import org.neo4j.driver.v1.Session;
import org.neo4j.driver.v1.StatementResult;
import org.neo4j.driver.v1.Transaction;
import org.neo4j.driver.v1.TransactionWork;
import org.neo4j.driver.v1.Value;
import org.neo4j.driver.v1.types.Node;

import static org.neo4j.driver.v1.Values.parameters;
/**
 * Tests basés sur le protocole BOLT+ROUTING (cluster causal uniquement)
 * @author sroussy
 *
 */
public class BoltRoutingCatalogueTest {

	// A changer par votre Url
	private static final String BOLT_URL ="bolt+routing://localhost:7689";
	
	// A changer par votre utilisateur
	private static final String USER ="neo4j";
	
	// A changer par votre mot de passe
	private static final String PASSWORD ="neo4j";
	
	Driver driver;
	
	/**
	 * Instanciation du driver BOLT
	 * @throws IOException
	 */
	@Before
    public void setup() throws IOException {
		final AuthToken token = AuthTokens.basic(USER, PASSWORD);
		driver = GraphDatabase.driver(BOLT_URL,token);
    }
	
	/**
	 * Test les transactions implicites
	 */
	@Test
	public void testImplicitTransaction (){
		try(Session sessionBolt = driver.session()){
			sessionBolt.run( "MERGE (p:Personne {nom: $nom})", parameters( "nom", "Nicolas" ) );
		}
		
		
		try(Session sessionBolt = driver.session()){
			final StatementResult sr = sessionBolt.run( "MATCH (p:Personne {nom: $nom}) RETURN p", parameters( "nom", "Nicolas" ) );
			while (sr.hasNext()){
				Record ligne = sr.next();
				Value colonneP = ligne.get("p");
				Node noeud = (Node) colonneP.asNode();
				Value valeur = noeud.get("nom");
		 		System.out.println(valeur.asString());
				
				Assert.assertEquals("Nicolas", noeud.get("nom").asString());
			}
		}
		
		
	}
	
	/**
	 * Test les transactions explicites
	 */
	@Test
	public void testExplicitTransaction (){
		String bookmark ="";
		
		try(Session sessionBolt = driver.session()){
			try (Transaction tx = sessionBolt.beginTransaction()){
				tx.run( "MERGE (p:Personne {nom: $nom})", parameters( "nom", "Nicolas" ) );
				tx.success();
				
				bookmark = sessionBolt.lastBookmark();
				
			}
		
		}
		
		
		try(Session sessionBolt = driver.session(bookmark)){
			try (Transaction tx = sessionBolt.beginTransaction()){
				final StatementResult sr = tx.run( "MATCH (p:Personne {nom: $nom}) RETURN p", parameters( "nom", "Nicolas" ) );
				while (sr.hasNext()){
					Record ligne = sr.next();
					Value colonneP = ligne.get("p");
					Node noeud = (Node) colonneP.asNode();
					Value valeur = noeud.get("nom");
			 		System.out.println(valeur.asString());
					
					Assert.assertEquals("Nicolas", noeud.get("nom").asString());
				}
				
			}
			
		}
		
		
	}
	
	/**
	 * Test les fonctions transactionnelles avec utilisation du Bookmark
	 */
	@Test
	public void testTransactionFunctions (){
		String bookmark ="";
		final String nom="Patrick";
		try(Session sessionBolt = driver.session()){
			sessionBolt.writeTransaction(new TransactionWork<Integer>() {
				@Override
	            public Integer execute( Transaction tx )
	            {					
	                return creerPersonnne(tx, nom );
	            }
			});
			bookmark = sessionBolt.lastBookmark();
			System.out.println(bookmark);
		}		
		
		
		try(Session sessionBolt = driver.session(bookmark)){
			Node noeud = sessionBolt.readTransaction(new TransactionWork<Node>() {
				@Override
	            public Node execute( Transaction tx )
	            {
	                return retrouvePersonne(tx, nom);
	            }
			});
			System.out.println(noeud.get("nom").asString());
			Assert.assertEquals("Nicolas", noeud.get("nom").asString());
		
		}
	}
	
	/**
	 * Créé un noeud personne dans le graphe
	 * @param tx
	 * @param nom
	 * @return
	 */
	private static int creerPersonnne( Transaction tx, String nom )
	{
	    tx.run( "MERGE (p:Personne {nom: $nom})", parameters( "nom", nom ) );
	    return 1;
	}
	
	/**
	 * Retrouve un noeud personne dans le graphe
	 * @param tx
	 * @param nom
	 * @return
	 */
	private static Node retrouvePersonne( Transaction tx, String nom )
	{
		Node node = null;
		final StatementResult sr = tx.run( "MATCH (p:Personne {nom: $nom}) RETURN p", parameters( "nom", nom ) );
		if (sr.hasNext()){
			node = sr.next().get("p").asNode();
		}
	    return node;
	}
	
	
	/**
	 * Relâche le driver BOLT
	 */
    @After
    public void closeDriver() {
        if (driver != null) driver.close();
    }
}
